"""Sequence Modules used in president."""
import logging
import os
import tempfile
import numpy as np

logger = logging.getLogger(__name__)


def to_valid_upper(dna_seq):
    """
    Transform input sequence to valid output sequence by removing all non-ACGT characters with Ns.

    Parameters
    ----------
    dna_seq, str
              DNA sequence

    Returns
    -------
    str, fixed DNA sequence
    """
    valid_nt = set(list("acgtACGT"))
    return "".join([i.upper() if i in valid_nt else "N" for i in dna_seq])


def preprocess(multifasta, suffix):
    """Remove white spaces from IDs in a FASTA file.

       Removes trailing spaces and newlines.

       Converts sequences to upper case.

       Concatenates n fasta files received as *.fasta

       Replaces - with . for pblat.

    Parameters
    ----------
    fasta : str/[str]
        FASTA file(s) to preprocess.
    suffix : str
        Suffix for the temporary output file.

    Returns
    -------
    processed FASTA file.

    """
    # Create temp file to save output
    _, output = tempfile.mkstemp(suffix=suffix)

    # Source file(s)
    source = []

    # If input is a string then convert it to a list
    if isinstance(multifasta, str):
        multifasta = [multifasta]

    # concat multiple input files
    with open(output, "w") as fout:
        for fasta in multifasta:
            logger.info(f"Preprocessing file: {fasta} -> {output}")
            with open(fasta, "r") as fin:
                for line in fin:
                    # removes trailing spaces and linebreaks
                    line = line.strip()
                    # replaces whitespaces with a placeholder %space% to avoid errors
                    # needs to be undone later in the code
                    if len(line) > 0 and line.startswith('>'):
                        source.append(os.path.basename(fasta))
                        fout.write(line.replace(" ", "%space%") + "\n")
                    else:
                        # convert sequence to upper case
                        line = line.upper()
                        # replaces gaps
                        line = line.replace('-', '.')
                        fout.write(line + "\n")
    return output, np.array(source)



"""Init module for president."""
import os
from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

files = os.listdir(os.path.dirname(__file__))
files.remove('__init__.py')
__all__ = [f[:-3] for f in files if f.endswith(".py")]



"""Module for writing and organizing results."""
import logging
import os
from datetime import datetime

import numpy as np
import screed

from president import alignment
from president import sequence

logger = logging.getLogger(__name__)

PSL_LABELS = alignment.PSL_LABELS

COL_ORDER = ["query_name", "reference_name", "file_in_query", "file_in_ref",
             "ACGT Nucleotide identity", "ACGT Nucleotide identity (ignoring Ns)",
             "ACGT Nucleotide identity (ignoring non-ACGTNs)", "qc_all_valid",
             "qc_is_empty_query", "qc_post_align_pass_threshold", "qc_post_aligned",
             "qc_post_aligned_all_valid", "qc_valid_length", "qc_valid_nucleotides",
             "qc_valid_pass_nthreshold", "acgt_bases", "iupac_bases",
             "non_iupac_bases", "N_bases", "length_query", "length_reference",
             "LongestNGap", "Matches", "Mismatches", "query_description",
             "query_index", "Date"]


def init_metrics(n_seqs, extend_cols=False, metrics_df=None, store_alignment=False):
    """
    Init output data structure.

    Parameters:
    ----------
    n_seqs: int,
            number of sequences to process / store meta values
    extend_cols: bool,
                if True, extends the columns to presidents output format.
                Needed for consistent writing in cases where no sequences were aligned.
    metrics_df:
        Potential Dataframe to be extended (default: None). Must be used with extend_cols.

    Returns:
    -------
    np structured array / pandas dataframe
    """
    statistic_values = np.dtype([("query_description", "object"),
                                 ("query_name", "object"),
                                 ("query_index", "object"),
                                 ("acgt_bases", np.uint32),
                                 ("iupac_bases", np.uint32),
                                 ("non_iupac_bases", np.uint32),
                                 ("N_bases", np.uint32),
                                 ("length_query", np.uint32),
                                 ("LongestNGap", np.uint32)])
    stats_ar = np.zeros(n_seqs, dtype=statistic_values)

    if extend_cols:
        # extend output data to be equal to the standard output
        needed_cols = \
            ["ACGT Nucleotide identity", "ACGT Nucleotide identity (ignoring Ns)",
             "ACGT Nucleotide identity (ignoring non-ACGTNs)", "Matches", "Mismatches",
             "qc_post_align_pass_threshold", "qc_post_aligned",
             "qc_post_aligned_all_valid", "reference_name", "length_reference"]
        if store_alignment:
            needed_cols += \
                ["PSL_"+c for c in PSL_LABELS]

        for coli in needed_cols:
            if "qc_post" in coli:
                if metrics_df.shape[0] == 0:
                    # If fasta file is empty then create a single row
                    metrics_df[coli] = [False]
                else:
                    # Otherwise create multiple rows for all invalid sequences
                    metrics_df[coli] = metrics_df.shape[0] * [False]
            else:
                metrics_df[coli] = np.nan
        metrics_df['Date'] = datetime.today().strftime('%Y-%m-%d')
        return metrics_df

    else:
        return stats_ar


def get_filename(file):
    """
    Extract filename of the file.

    Parameters
    ----------
    file : str
        file location.
    Returns
    -------
    str, filename without extension
    """
    return os.path.splitext(os.path.basename(file))[0]


def write_fasta(fileobj, seq, format_sequence=False, format_pblat=False):
    """
    Write fasta sequence to file.

    Parameters
    ----------
    fileobj: obj,
        file object (opened)
    seq: seq object,
        screed sequence object
    format_sequence: bool,
        If True, transform sequence to valid upper case sequence (default: False)
    Returns
    -------
        None
    """
    # pblat cuts away the description, fiddle around this issue by having the " " decoded as %space%
    if not format_pblat:
        sequence_header = f">{seq.name} {seq.description}".replace("%space%", " ")
    else:
        sequence_header = f">{seq.name}{seq.description}"
    fileobj.write(sequence_header.rstrip()+"\n")
    if format_sequence:
        # make sure to only store upper case, ACGT symbols. replace all others with "N"s
        fileobj.write(sequence.to_valid_upper(seq.sequence) + "\n")
    else:
        fileobj.write(f"{seq.sequence}\n")


def write_sequences(query, metrics, out_dir, evaluation, write_mode="w"):
    """
    Write sequences from the query into valid / invalid FASTA files based on metrics.

    Parameters
    ----------
    query: str,
        file location of the query FASTA
    metrics: dataframe,
            pandas dataframe filled with valid and aligned columnns
    out_dir: str,
            Output directory to store FASTAs

    Returns
    -------
        None
    """
    prefix = os.path.basename(out_dir)
    if evaluation != "all_invalid":
        # valid ids were aligned and pass the qc
        valid_ids = set(metrics[metrics["qc_post_aligned_all_valid"]]["query_name"].values)
    else:
        valid_ids = set()

    # split valid and invalid file writers
    valid_name = os.path.join(os.path.dirname(out_dir), f"{prefix}valid.fasta")
    valid_fout = open(valid_name, write_mode)
    invalid_name = os.path.join(os.path.dirname(out_dir), f"{prefix}invalid.fasta")
    invalid_fout = open(invalid_name, write_mode)

    # iterate over query and split into valid / invalid fasta
    nvalid = 0
    ninvalid = 0
    with screed.open(query) as seqfile:
        for seq in seqfile:
            if seq.name.replace("%space%", " ") in valid_ids:
                nvalid += 1
                write_fasta(valid_fout, seq, format_sequence=True)
            else:
                write_fasta(invalid_fout, seq)
                ninvalid += 1

    logger.info(f"Valid (n={nvalid}) sequences written to: {valid_name}")
    logger.info(f"Invalid (n={ninvalid}) sequences written to: {invalid_name}")
    valid_fout.close()
    invalid_fout.close()



"""President's main module.

Given a reference and a query sequence (which can be fragmented), calculate
pairwise nucleotide identity with respect to the reference sequence.

Usage:
-----
president --query tiny_test.masked_consensus.fasta \
    --reference NC_045512.2.fasta -x 0.9 -p output/report -t 4

Notes:
-----
- nextstrain uses a quality threshold of < 3000 non-canonical nucleotides
- Ns in the query are treated as mismatches, uncomment the corresponding line
(below) to ignore Ns

Requirements:

- conda install -y -c bioconda python=3.8 pblat screed pandas

ANI definition:

- https://pubmed.ncbi.nlm.nih.gov/17220447/

Spec:
-----
> Sequence identity to a reference genome sequence (e.g. NC_045512.2) should
be calculated at the nucleotide level relative to the entire length of the
reference. Only informative nucleotides (A,T,G,C) are considered identical
to each other.
"""
# authors:
# RKI MF1;  Martin Hoelzer with great initial help of @phiweger (UKL Leipzig)
# HPI;      Fabio Malcher Miranda, Sven Giese, Alice Wittig
import argparse
import os
from shutil import which
import sys
from datetime import datetime
import pandas as pd
import logging
from president import alignment, __version__, statistics, writer, sequence

logger = logging.getLogger(__name__)


def is_available(name="pblat"):
    """
    Check whether `name` is on PATH and marked as executable.

    Parameters
    ----------
    name : str
        tool name executable (default: pblat).

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    # tool installed
    if not which(name):
        raise ValueError(f'{name} not on PATH or marked as executable.')


def aligner(reference_in, query_in_raw, path_out, prefix_out="",
            id_threshold=0.9, n_threshold=0.05,
            threads=4, store_alignment=False, verbose=False):  # pragma: no cover
    """
    Align query to the reference and extract qc metrics.

    Parameters
    ----------
    reference_in : str
        reference FASTA location
    query_in_raw : str / list
        query FASTA location(s).
    path_out : str
        Path to be used to store the results.
    prefix_out : str
        File prefix to be used for each output file.
    id_threshold : float, optional
        Identity threshold after aligment that must be achieved. The default is 0.9.
    n_threshold: float, optional
        Percentage of allowed Ns, sequences with hier N% will be rejected
    threads : int, optional
        Number of threads to use. The default is 4.
    store_alignment : bool, optional
        Should we store the results of the alignements in the qc metrics table? default is False
    verbose: bool,
            if True, print logging information to screen
    Returns
    -------
    datframe,
        result metrics
    """
    # handle path / prefix input
    out_dir = os.path.abspath(path_out)
    preloginfo = ""
    if not os.path.exists(out_dir):
        preloginfo = "Creating output directory..."
        os.makedirs(out_dir, exist_ok=True)

    # create logger
    logger = logging.getLogger('president')
    logger.setLevel(logging.DEBUG)
    # create console handler and set level to debug
    ch = logging.FileHandler(os.path.join(path_out, f"{prefix_out}president_logger.log"), "w")
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    logger.addHandler(ch)

    if not verbose:
        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(logging.DEBUG)
        sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
        logger.addHandler(sh)

    logger.info("Starting Time: {}".format(datetime.now().strftime("%H:%M:%S")))
    logger.info("Using president version: {}".format(__version__))
    logger.info(f"Writing files to: {path_out}")
    logger.info(f"Using the prefix: {prefix_out}_* to store results.")

    logger.info(preloginfo)
    logger.info(f"Query: {query_in_raw}")
    logger.info(f"Reference: {reference_in}")
    # Files exist?
    if not os.path.isfile(reference_in):
        logger.error(f"Reference {reference_in} does not exist.")
        raise FileNotFoundError(reference_in)

    # make sure input is iterable
    if not isinstance(query_in_raw, list):
        query_in_raw = [query_in_raw]

    collect_dfs = []
    logger.info("Starting president processing")
    # preprocess fasta files
    query_tmp, query_source = sequence.preprocess(query_in_raw, "_query.fasta")
    reference_tmp, _ = sequence.preprocess(reference_in, "_reference.fasta")

    prefix = prefix_out
    assert os.path.isfile(query_tmp)

    # check reference fasta
    _ = statistics.count_sequences(reference_tmp)
    query_valid = statistics.count_sequences(query_tmp, "query")

    # check query data
    if query_valid:
        summary_stats_query = statistics.summarize_query(query_tmp)
        statistics.qc_check(reference_tmp, summary_stats_query, id_threshold=id_threshold,
                            n_threshold=n_threshold)

        # perform initial sequence check
        query_valid_tmp, evaluation, invalid_ids = \
            statistics.split_valid_sequences(query_tmp, summary_stats_query)

        logger.info(f"Performing alignment with valid sequences (excluding {len(invalid_ids)}).")
    else:
        # make sure mock dataframe looks like regular one
        evaluation = "all_invalid"
        summary_stats_query = pd.DataFrame(writer.init_metrics(0))
        summary_stats_query = \
            summary_stats_query.assign(**{'qc_all_valid': [], 'qc_valid_length': [],
                                          'qc_valid_nucleotides': [], 'qc_valid_number_n': [],
                                          'qc_valid_pass_nthreshold': []})

    # align sequences if more than 1 sequence passes the initial qc
    if evaluation != "all_invalid":
        alignment_file = alignment.pblat(threads, reference_tmp, query_valid_tmp)
        # parse statistics from file
        metrics = statistics.nucleotide_identity(alignment_file, summary_stats_query,
                                                 id_threshold, store_alignment)
        metrics["qc_is_empty_query"] = False
    else:
        # if no sequences are there to be aligned, create a pseudooutput that looks
        # exactly as the aligned output
        metrics = writer.init_metrics(1, extend_cols=True, metrics_df=summary_stats_query)
        metrics["qc_is_empty_query"] = True
    # store sequences
    writer.write_sequences(query_tmp, metrics, os.path.join(out_dir, f"{prefix}"), evaluation)

    # store reference data
    if len(query_source) > 0:
        metrics["file_in_query"] = query_source[metrics["query_index"].astype(int)]
    else:
        metrics["file_in_query"] = 'NaN'

    metrics["file_in_ref"] = os.path.basename(reference_in)

    # pretify output
    metrics = metrics[metrics.columns.sort_values()]
    # putting the columns with PSL prefix at the end
    PSL_columns = metrics.filter(regex="^PSL_").columns.to_list()
    metrics = metrics[writer.COL_ORDER + PSL_columns]
    metrics.to_csv(os.path.join(out_dir, f"{prefix}report.tsv"), index=False, sep='\t')

    # remove temporary files
    if evaluation != "all_invalid":
        os.remove(alignment_file)

    os.remove(query_tmp)
    os.remove(reference_tmp)
    logger.info(metrics)
    logger.info(metrics.shape)
    collect_dfs.append(metrics)

    # if there are more input files to iterate from, concat results
    metrics_all = pd.concat(collect_dfs)

    logger.info("president finished from command:")
    logger.info(f"Call:'president -r {reference_in} -q {query_in_raw} -p {path_out} "
                f"-f {prefix_out} -x {id_threshold} -n {n_threshold} -a {store_alignment} "
                f"-t {threads}'")
    return metrics_all


def main():  # pragma: no cover
    """
    Presidents main function for sequence alignment.

    Returns
    -------
    None.

    """
    parser = argparse.ArgumentParser(description='Calculate pairwise nucleotide identity.')
    parser.add_argument('-r', '--reference', required=True, help='Reference genome.')
    parser.add_argument('-q', '--query', required=True, help='Query genome(s).', nargs="+")
    parser.add_argument('-x', '--id_threshold', type=float, default=0.9,
                        help='ACGT nucleotide identity threshold after alignment (percentage). '
                             'A query sequence is reported as valid based on this threshold '
                             '(def: 0.9)')
    parser.add_argument('-n', '--n_threshold', type=float, default=0.05,
                        help='A query sequence is reported as valid, if the percentage of Ns '
                             'is smaller or equal the threshold (def: 0.05)')
    parser.add_argument('-t', '--threads', type=int, default=4,
                        help='Number of threads to use for pblat.')
    parser.add_argument('-p', '--path', required=True,
                        help='Path to be used to store results and FASTA files.')
    parser.add_argument('-f', '--prefix', required=False, default="",
                        help='Prefix to be used t store results in the path')
    parser.add_argument('-a', '--store_alignment',
                        required=False, action="store_true",
                        help='add query alignment columns (PSL format)')
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument('-e', '--quite', dest="verbose", action="store_true", default=False,
                        help="Print log messages also to the screen (False)", required=False)
    args = parser.parse_args()

    aligner(args.reference, args.query, args.path, args.prefix, args.id_threshold,
            args.n_threshold, args.threads, args.store_alignment, args.verbose)


if __name__ == "__main__":
    main()



"""Alignment Modules used in president."""
import subprocess
import tempfile
import logging
import pandas as pd

logger = logging.getLogger(__name__)

PSL_LABELS = ['Matches', 'Mismatches', 'RepMatch', 'Ns', 'QGapCount',
              'QGapBases', 'TGapCount', 'TGapBases', 'Strand',
              'QName', 'QSize', 'QStart', 'QEnd', 'TName', 'TSize',
              'TStart', 'TEnd', 'BlockCount', 'BlockSizes',
              'QStarts', 'TStarts']


def pblat(threads, reference, query):
    """Perform blat alignment.

    Parameters
    ----------
    threads : int
        Number of threads to use.
    reference : str
        file location of reference FASTA.
    query : str
        file location of query FASTA.

    Returns
    -------
    alignments.

    """
    _, alignments = tempfile.mkstemp(suffix="_pblat.tsv")
    cmd = f'pblat -threads={threads} {reference} {query} {alignments}'
    logger.info(f'Running pblat with: {cmd}')
    _ = subprocess.check_output(cmd, shell=True)
    logger.info('Finished pblat.')
    return alignments


def parse_alignment(alignment_file):
    """
    Parse a given alignment file into a pandas dataframe.

    Parameters
    ----------
    alignment_file : str
        location of the alignment file from pblat.

    Returns
    -------
    None.

    """
    # Read the alignment(s) with pandas
    # Pandas can be replaced with split to reduce dependencies
    try:
        alignments = pd.read_csv(alignment_file,
                                 header=None, sep='\t', skiprows=5)
        alignments.columns = PSL_LABELS
        alignments.QName = alignments.QName.astype(str)

    except pd.errors.EmptyDataError:
        logger.error("Error reading pblat output. Perhaps it did not align anything (pandas).")
        return pd.DataFrame()

    except Exception:
        logger.error("Error reading pblat output. Perhaps it did not align anything (exception).")
        return pd.DataFrame()
    return alignments



"""setup file for the project."""
# code gratefully take from https://github.com/navdeep-G/setup.py

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
import versioneer
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'pypresident'
DESCRIPTION = \
    'PaiRwisE Sequence IDENtiTy. Calculate pairwise nucleotide identity' + \
    'with respect to a reference sequence.'

# URL = 'https://gitlab.com/RKIBioinformaticsPipelines/president'
# URL_DOKU = "https://gitlab.com/RKIBioinformaticsPipelines/president
URL_GITHUB = "https://gitlab.com/RKIBioinformaticsPipelines/president"
URL_ISSUES = "https://gitlab.com/RKIBioinformaticsPipelines/president/-/issues"
EMAIL = 'HoelzerM@rki.de'
AUTHOR = 'Martin Hölzer'
REQUIRES_PYTHON = '>=3.8.0'
KEYWORDS = ["sequence alignment", "corona", "pblat", "SARS-COV2"]
RKI_SOFTWARE = "https://gitlab.com/RKIBioinformaticsPipelines"
# What packages are required for this module to be executed?
REQUIRED = ['pandas', 'screed', 'numpy']

# What packages are optional?
# 'fancy feature': ['django'],}
EXTRAS = {}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's _version.py module as a dictionary.
about = {}
# project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
project_slug = "president"
# with open(os.path.join(here, project_slug, '_version.py')) as f:
#    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Print things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        """Init options."""
        pass

    def finalize_options(self):
        """Finalize method."""
        pass

    def run(self):
        """Run method."""
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(
            sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    project_urls={
        "Bug Tracker": URL_ISSUES,
        "Source Code": URL_GITHUB,
        # "Documentation": URL_DOKU,
        # "Homepage": URL,
        "Related Software": RKI_SOFTWARE},
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],

    # 'mycli=mymodule:cli'
    entry_points={'console_scripts': ["president=president.__main__:main"],
                  },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='Apache License 2.0',
    keywords=KEYWORDS,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)
#    # $ setup.py publish support.
#    cmdclass={
#        'upload': UploadCommand,
#    },
